/* 

Copyright (C) 2003-2004 Yves Schmid (www.garagecube.com)

Author(s): Yves Schmid

This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     GNU General Public License for more details.

You should have received a copy of the GNU General Public Licensealong with this program; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#import "CosmicPalette.h"
#import "CosmicDocument.h"
#import "CosmicOpenGLView.h"

#include "OMediaCanvas.h"
#include "OMediaFilePath.h"
#include "OMediaString.h"
#include "OMediaFileStream.h"
#include "OMediaMoveableMem.h"
#include "OMediaMemStream.h"

@implementation CosmicPalette

static CosmicPalette *mainPtr;

-(OMediaCanvas*)readCanvas:(string*)pathname
{
	OMediaFilePath		fpath;
	OMediaFileStream	stream;
	OMediaCanvas		*canvas;

	fpath.setpath(*pathname);
	stream.setpath(&fpath); 
	
	stream.open(omcfp_Read,false,false);
	stream.setposition(0);

	canvas = new OMediaCanvas;

	stream>>canvas;	
	stream.close();

	return canvas;
}

-(id)init
{
	int i;
	string			str;

	self = [super init];
	mainPtr = self;

	currentTool = cosmic_Draw;
	
	[[NSColorPanel sharedColorPanel] setShowsAlpha:YES];

	// Read brushes
	
	for(i=0;i<COSMIC_NBRUSHES;i++)
	{
		switch(i/5)
		{
			case 0: str = "brusha"; break;
			case 1: str = "brushb"; break;
			case 2: str = "brushc"; break;
			case 3: str = "brushd"; break;		
			case 4: str = "brushe"; break;
		}
		
		str += omd_L2STR((i%5)+1);
		str += ".png";
		
		brushes[i] = [self readCanvas:&str];
	}
			
	return self;
}

-(void)dealloc
{
	int i;
	
	for(i=0;i<COSMIC_NBRUSHES;i++)
	{
		delete brushes[i];
	}
	
	for(i=0;i<COSMIC_NIMAGES;i++)
	{
		delete imageCanvas[i];
	}
	
	delete currentBrush;
	delete currentBrushShadow;

	[super dealloc];
}

+(CosmicPalette*)getMain
{
	return mainPtr;
}

-(int)findCurrentBrush
{  
	return [brushMatrix selectedRow]*5 + [brushMatrix selectedColumn];
}

-(void)fillBrushColor:(OMediaFARGBColor*)argb
{
	if (!currentBrush || !currentBrushShadow) return;

	unsigned long sa,sr,sg,sb;
	unsigned long a,r,g,b;

	a = (unsigned long)(argb->alpha*256);
	r = (unsigned long)(argb->red*256);
	g = (unsigned long)(argb->green*256);
	b = (unsigned long)(argb->blue*256);
	
	currentBrush->lock(omlf_Write|omlf_Read);
	currentBrushShadow->lock(omlf_Write|omlf_Read);

	unsigned long *shpixels = (unsigned long*)currentBrushShadow->get_pixels();
	unsigned long *pixels = (unsigned long*)currentBrush->get_pixels();
	unsigned long *last_pixels = pixels + (currentBrush->get_width() * currentBrush->get_height());
	unsigned long pix;

	for(;pixels!=last_pixels;pixels++,shpixels++)
	{
		pix = *pixels;
		sr = pix>>24;
		sg = (pix>>16)&0xFF;
		sb = (pix>>8)&0xFF;
		sa = (pix)&0xFF;

		sa = (sa * a)>>8;
		sr = (sr * r)>>8;
		sg = (sg * g)>>8;
		sb = (sb * b)>>8;
		
		*pixels = (sr<<24)|(sg<<16)|(sb<<8)|(sa);
		*shpixels = sa>>1;
	}
	
	currentBrush->unlock();
	currentBrushShadow->unlock();

}

-(void)initCurrentBrushFromImage
{
	int p;

	p = [imageSlider intValue];
	if (p<0) p=0;
	else if (p>=COSMIC_NIMAGES) p=COSMIC_NIMAGES-1;

	if (images[p]==NULL)
	{
		currentBrush->create(1,1);  // Empty brush
		currentBrush->fill_alpha(0);
	}
	else
	{
		if (imageCanvas[p]!=NULL)
		{	
			currentBrush->create(imageCanvas[p]->get_width(),imageCanvas[p]->get_height(),imageCanvas[p]);
		}
		else
		{
			NSImage				*srcimage = images[p],*image;
			NSBitmapImageRep	*bitmap;
			NSRect				rect;

			int width = (int)[srcimage size].width,
				height = (int)[srcimage size].height;

			rect.origin.x = 0;
			rect.origin.y = 0;
			rect.size.width = width;
			rect.size.height = height;

			image = [[NSImage alloc] initWithSize:[srcimage size]];
			[image lockFocus];
			[srcimage drawAtPoint:rect.origin fromRect:rect operation:NSCompositeCopy fraction:1.0];
			bitmap = [[NSBitmapImageRep alloc] initWithFocusedViewRect:NSMakeRect (0.0f, 0.0f, width,height)];
			[image unlockFocus];

			currentBrush->create(width,height);
			currentBrush->lock(omlf_Write|omlf_Read);
			
			unsigned char *s,*bs,*d,*bd;
			int				bitsPerPixel;
			
			bitsPerPixel = [bitmap bitsPerPixel];
			bs = [bitmap bitmapData];
			bd = (unsigned char*)currentBrush->get_pixels();
			
			for(int y=0;y<height;y++)
			{
				s = bs + [bitmap bytesPerRow]*y;
				d = bd + (width<<2)*y;

				switch(bitsPerPixel)
				{
					case 8:		
					for(int x=0;x<width;x++)
					{
						*(d++) = *s;
						*(d++) = *s;
						*(d++) = *s;
						*(d++) = 0xFF;
						s++;		
					}
					break;

					case 16:		
					for(int x=0;x<width;x++)
					{
						*(d++) = ((*s)>>10)&(0x1F);
						*(d++) = ((*s)>>5)&(0x1F);
						*(d++) = (*s)&(0x1F);
						*(d++) = 0xFF;
						s+=2;		
					}
					break;
					
					case 24:		
					for(int x=0;x<width;x++)
					{
						*(d++) = *(s++);
						*(d++) = *(s++);
						*(d++) = *(s++);
						*(d++) = 0xFF;
					}
					break;

					case 32:		
					for(int x=0;x<width;x++)
					{
						*(d++) = *(s++);
						*(d++) = *(s++);
						*(d++) = *(s++);
						*(d++) = *(s++);
					}
					break;
				}
			}

			currentBrush->unlock();
		
			[bitmap release];
			[image release];
			
			imageCanvas[p] = new OMediaCanvas;
			imageCanvas[p]->create(currentBrush->get_width(),currentBrush->get_height(),currentBrush);
		}
	}
}

-(void)buildCurrentBrush
{
	if (!currentBrush) 
	{
		currentBrush = new OMediaCanvas;
		currentBrushShadow = new OMediaCanvas;
	}

	// Find current brush

	int p = [self findCurrentBrush];
	if (p<0 || p>=COSMIC_NBRUSHES) return;

	if ([[[brushTabView selectedTabViewItem] identifier] isEqualToString:@"standard"])
	{
		currentBrush->create(brushes[p]->get_width(),brushes[p]->get_height(),brushes[p]);
	}
	else
	{
		[self initCurrentBrushFromImage];
	}

	currentBrushShadow->create(currentBrush->get_width(),currentBrush->get_height());	


	OMediaFARGBColor	startARGB,endARGB,finalARGB;

    [[[startColorWell color] colorUsingColorSpaceName:@"NSDeviceRGBColorSpace"]  getRed:&startARGB.red green:&startARGB.green blue:&startARGB.blue alpha:&startARGB.alpha];
    [[[endColorWell color] colorUsingColorSpaceName:@"NSDeviceRGBColorSpace"]  getRed:&endARGB.red green:&endARGB.green blue:&endARGB.blue alpha:&endARGB.alpha];

	finalARGB.alpha = startARGB.alpha*(1.0f-currentColorState) + endARGB.alpha*(currentColorState);
	finalARGB.red = startARGB.red*(1.0f-currentColorState) + endARGB.red*(currentColorState);
	finalARGB.green = startARGB.green*(1.0f-currentColorState) + endARGB.green*(currentColorState);
	finalARGB.blue = startARGB.blue*(1.0f-currentColorState) + endARGB.blue*(currentColorState);

	[self fillBrushColor:&finalARGB];
	
	if (currentBrush->get_width()>128 || 
		currentBrush->get_height()>128)
	{
		if ([toolConnectedDrawButton isEnabled])
		{
			[toolConnectedDrawButton setState:NSOffState];
			[toolConnectedDrawButton setEnabled:NO];
			[toolDrawButton setState:NSOnState];
			currentTool = cosmic_Draw;
		}
	}
	else if (![toolConnectedDrawButton isEnabled])
	{
		[toolConnectedDrawButton setEnabled:YES];
	}
}

-(void)rebuildWithColorState:(float)colorInterpolateState
{
	if (currentColorStateInt==(int)(colorInterpolateState * (float)0xFF)) return;
	currentColorState = colorInterpolateState;
	currentColorStateInt = (int)(colorInterpolateState * (float)0xFF);

	[self buildCurrentBrush];
}

-(OMediaCanvas*)getCurrentBrush
{
	if (!currentBrush) [self buildCurrentBrush];
	
	return currentBrush;
}

-(IBAction)brushChanged:(id)sender
{
	[self buildCurrentBrush];
}

-(IBAction)colorChanged:(id)sender
{
	[self buildCurrentBrush];
}

-(IBAction)imageChanged:(id)sender
{
	int p;
	
	p = [imageSlider intValue];
	if (p<0) p=0;
	else if (p>=COSMIC_NIMAGES) p=COSMIC_NIMAGES-1;
	
	if (images[p]) [images[p] release];
	images[p] = [[imageView image] retain];

	delete imageCanvas[p];
	imageCanvas[p] = NULL;

	dirtyImages[p] = YES;

	[self buildCurrentBrush];
}

-(IBAction)currentImagePosChanged:(id)sender
{
	int p;
	
	p = [imageSlider intValue];
	if (p<0) p=0;
	else if (p>=COSMIC_NIMAGES) p=COSMIC_NIMAGES-1;

	[imageTextField setStringValue:[NSString stringWithFormat: @"Image (%d/%d)",p+1,COSMIC_NIMAGES]];
	[imageView setImage:images[p]];

	[self buildCurrentBrush];
}


-(IBAction)showHelp:(id)sender
{
	NSString *path = [[NSBundle mainBundle] bundlePath];

	path = [path stringByDeletingLastPathComponent];
	path = [path stringByAppendingPathComponent:@"Documentation.rtfd"];

	[[NSWorkspace sharedWorkspace] openFile:path];	
}

- (void)tabView:(NSTabView *)tabView didSelectTabViewItem:(NSTabViewItem *)tabViewItem
{
	[self buildCurrentBrush];
}

- (void)controlTextDidChange:(NSNotification *)aNotification
{
	id obj = [aNotification object];

	if (obj==rotationSpeedTextField || obj==fpsTextField)
	{
		[rotationSpeedSlider setDoubleValue:([rotationSpeedTextField doubleValue]/100)*1.5];	
		[fpsSlider setDoubleValue:([fpsTextField doubleValue]/100.0)];	

		[rotationSpeedSlider sendAction:@selector(updateRotationValues:) to:nil];
	}
	else if (obj==scrollSpeedTextField)
	{
		[scrollSpeedSlider setDoubleValue:([scrollSpeedTextField doubleValue]/100)];	
		[scrollSpeedSlider sendAction:@selector(updateScrollValues:) to:nil];
	}
	else if (obj==scaleSpeedTextField || obj==scaleStartTextField || obj==scaleEndTextField)
	{
		[scaleSpeedSlider setDoubleValue:([scaleSpeedTextField doubleValue]/100)];	
		[scaleSpeedSlider sendAction:@selector(updateScaleValues:) to:nil];
	}
	else if (obj==manuallySetXTextField) [manuallySetXTextField sendAction:@selector(updateManuallySetValues:) to:nil];
	else if (obj==manuallySetYTextField) [manuallySetYTextField sendAction:@selector(updateManuallySetValues:) to:nil];
	else if (obj==manuallySetAngleTextField) [manuallySetAngleTextField sendAction:@selector(updateManuallySetValues:) to:nil];
	else if (obj==manuallySetZoomTextField) [manuallySetZoomTextField sendAction:@selector(updateManuallySetValues:) to:nil];
}

-(void)resetAllCursors
{   
	NSArray			*array = [[NSDocumentController sharedDocumentController] documents];
	CosmicDocument  *document;

	for(int i=0;i<(int)[array count];i++)
	{
		document = (CosmicDocument*)[array objectAtIndex:i];
		if (!document) continue;
		
		[[document->mainView window] invalidateCursorRectsForView:document->mainView];
	}
}

-(IBAction)toolChanged:(id)sender
{
	if (sender==toolDrawButton) currentTool = cosmic_Draw;
	else if (sender==toolConnectedDrawButton) currentTool = cosmic_ConnectedDraw;
	else if (sender==toolRotateButton) currentTool = cosmic_Rotate;
	else if (sender==toolScrollButton) currentTool = cosmic_Scroll;
	else if (sender==toolScaleButton) currentTool = cosmic_Scale;
	
	if ([sender state]==NSOffState)
			[sender setState:NSOnState];
	else
	{
		if (sender!=toolDrawButton) [toolDrawButton setState:NSOffState];
		if (sender!=toolConnectedDrawButton) [toolConnectedDrawButton setState:NSOffState];
		if (sender!=toolRotateButton) [toolRotateButton setState:NSOffState];
		if (sender!=toolScrollButton) [toolScrollButton setState:NSOffState];	
		if (sender!=toolScaleButton) [toolScaleButton setState:NSOffState];	
	}
	
	if (currentTool==cosmic_Draw || currentTool==cosmic_ConnectedDraw)
		[toolInfoTabView selectTabViewItemAtIndex:0];	
	else
		[toolInfoTabView selectTabViewItemAtIndex:1];	
	
	[self resetAllCursors];
}



@end

